package com.ibm.tools.attach;

/*
 * Licensed Materials - Property of IBM,
 *     Copyright IBM Corp. 2009, 2010  All Rights Reserved
 */

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import com.ibm.tools.attach.javaSE.Permissions;
import com.ibm.tools.attach.spi.AttachProvider;

/**
 * Handles the initiator end of an attachment to a target VM.
 *
 * This class is analogous to and intended to be compatible with
 * com.sun.tools.attach.VirtualMachine.
 *
 * @see "http://w3.hursley.ibm.com/java/docs/java7/jdk/api/attach/spec/com/sun/tools/attach/VirtualMachine.html"
 *
 */
public abstract class VirtualMachine {

	private final AttachProvider provider;

	/**
	 * ID of the target VM
	 */
	private final String id;

	/**
	 * @param provider
	 *            AttachProvider
	 * @param id
	 *            of target VM
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	protected VirtualMachine(AttachProvider provider, String id) {
		this.provider = provider;
		this.id = id;
	}

	/**
	 * @return list of available virtual machines
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public static List<VirtualMachineDescriptor> list() {
		List<AttachProvider> providerList = AttachProvider.providers();
		List<VirtualMachineDescriptor> allVMs = new ArrayList<VirtualMachineDescriptor>();

		for (AttachProvider provider : providerList) {
			List<VirtualMachineDescriptor> vms = provider.listVirtualMachines();
			if (vms != null) {
				allVMs.addAll(vms);
			}
		}

		return allVMs;
	}

	/**
	 * @param id
	 *            ID of the target
	 * @return VirtualMachine object for attached VM
	 * @throws AttachNotSupportedException
	 *             if target does not respond
	 * @throws IOException
	 *             if control files are inaccessible.
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public static VirtualMachine attach(String id)
			throws AttachNotSupportedException, IOException {
		checkSecurity();
		if (null == id) {
			// K0554 = Virtual machine ID or display name is null
			throw new NullPointerException(com.ibm.oti.util.Msg.getString("K0554")); //$NON-NLS-1$
		}

		List<com.ibm.tools.attach.spi.AttachProvider> providers = com.ibm.tools.attach.spi.AttachProvider.providers();
		if ((null == providers) || providers.isEmpty()) {
			// K0534 = no attach providers found
			throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0534")); //$NON-NLS-1$
		}

		/* try each of the providers in order */
		for (com.ibm.tools.attach.spi.AttachProvider provider : providers) {
			try {
				VirtualMachine vm = provider.attachVirtualMachine(id);
				if (vm != null) {
					return vm;
				}
			} catch (AttachNotSupportedException e) {
			}
		}
		// K0542 = No provider for virtual machine id {0}
		throw new AttachNotSupportedException(com.ibm.oti.util.Msg.getString("K0542", id)); //$NON-NLS-1$
	}

	/**
	 * @param descriptor
	 *            VirtualMachineDescriptor for target
	 * @return VirtualMachine object for attached VM
	 * @throws AttachNotSupportedException
	 *             if target does not respond
	 * @throws IOException
	 *             if control files are inaccessible
	 * @throws SecurityException
	 *             The security manager (if installed) denies
	 *             AttachPermission("attachVirtualMachine")
	 * @throws NullPointerException
	 *             If the VirtualMachineDescriptor argument is null
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public static VirtualMachine attach(VirtualMachineDescriptor descriptor)
			throws AttachNotSupportedException, IOException {
		checkSecurity();
		if (null == descriptor) {
			// K0554 = Virtual machine ID or display name is null
			throw new NullPointerException(com.ibm.oti.util.Msg.getString("K0554")); //$NON-NLS-1$
		}

		AttachProvider provider = descriptor.provider();
		return provider.attachVirtualMachine(descriptor);
	}

	/**
	 * Disconnect from the target VM
	 *
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract void detach() throws IOException;

	/**
	 *
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public boolean equals(Object comparand) {

		if (comparand == null || !(comparand instanceof VirtualMachine)) {
			return false;
		}

		VirtualMachine otherVM = (VirtualMachine) comparand;
		return id().equals(otherVM.id());
	}

	/**
	 * @return agent properties maintained in the target
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract Properties getAgentProperties() throws IOException;

	/**
	 * @return properties of target VM
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract Properties getSystemProperties() throws IOException;

	/**
	 *
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	@Override
	public int hashCode() {
		return provider().hashCode() + id().hashCode();
	}

	/**
	 * @return id of target VM
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public final String id() {
		return id;
	}

	/**
	 * Convenience method: equivalent to calling loadAgent with empty options
	 * string
	 *
	 * @param agent
	 *            See loadAgent(String agent, String options)
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public void loadAgent(String agent) throws AgentLoadException,
			AgentInitializationException, IOException {
		loadAgent(agent, ""); //$NON-NLS-1$
	}

	/**
	 * load an agent per java.lang.instrument
	 *
	 * @param agent
	 *            path to an agent JAR file
	 * @param options
	 *            options to agentmain
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract void loadAgent(String agent, String options)
			throws AgentLoadException, AgentInitializationException,
			IOException;

	/**
	 * Convenience method: equivalent to calling loadAgentLibrary with empty
	 * options string
	 *
	 * @param agentLibrary
	 *            Name of the agent library, without "lib", ".so", etc.
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public void loadAgentLibrary(String agentLibrary)
			throws AgentLoadException, AgentInitializationException,
			IOException {
		loadAgentLibrary(agentLibrary, ""); //$NON-NLS-1$
	}

	/**
	 * @param agentLibrary
	 *            Name of the agent library, without "lib", ".so", etc.
	 * @param options
	 *            Argument to Agent_OnAttach
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract void loadAgentLibrary(String agentLibrary, String options)
			throws AgentLoadException, AgentInitializationException,
			IOException;

	/**
	 * Convenience method: equivalent to calling loadAgentPath with empty
	 * options string
	 *
	 * @param agentPath
	 *            Full path to filename of agent library
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public void loadAgentPath(String agentPath) throws AgentLoadException,
			AgentInitializationException, IOException {
		loadAgentPath(agentPath, ""); //$NON-NLS-1$
	}

	/**
	 * @param agentPath
	 *            Full path to filename of agent library
	 * @param options
	 *            See loadAgentLibrary
	 * @throws AgentLoadException
	 *             if target returns this exception
	 * @throws AgentInitializationException
	 *             if target returns this exception
	 * @throws IOException
	 *             if control files are inaccessible
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public abstract void loadAgentPath(String agentPath, String options)
			throws AgentLoadException, AgentInitializationException,
			IOException;

	/**
	 * @return AttachProvider set by the constructor
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public final AttachProvider provider() {
		return provider;

	}

	/**
	 *
	 * @note Public API, signature compatible with
	 *       com.sun.tools.attach.spi.AttachProvider.
	 */
	public String toString() {
		return getClass().getName() + "@" + //$NON-NLS-1$
				hashCode() + ";" + id(); //$NON-NLS-1$
	}

	private static void checkSecurity() {
		final SecurityManager securityManager = System.getSecurityManager();
		if (securityManager != null) {
			securityManager.checkPermission(Permissions.ATTACH_VM);
		}
	}

}
